@@ -38,3 +38,9 @@ pub use crate::{
     resources::{Res, ResMut, Resources},
     world::{Entity, World},
 };
+
+#[cfg(feature = "rayon")]
+mod rayon;
+
+#[cfg(feature = "rayon")]
+pub use crate::rayon::QueryParIter;
