@@ -7,7 +7,7 @@ use std::sync::atomic::{AtomicU32, Ordering};
 
 use crate::archetype::{Archetype, Comp, CompMut, TypeMetadataSet};
 
-/// The ECS world storing [Entities](Entity) and components.
+/// The world storing entities and their components.
 pub struct World {
     tag: u32,
     pub(crate) entities: Vec<EntityMetadata>,
@@ -178,14 +178,16 @@ impl World {
         let meta = &self.entities[ent.id as usize];
         assert_eq!(ent.gen, meta.gen, "Entity is stale");
 
-        let new_ty = if let Some(ty) = self.insert_map.get(&(meta.ty, TypeId::of::<B>())) {
+        let key = TypeId::of::<B>();
+
+        let new_ty = if let Some(ty) = self.insert_map.get(&(meta.ty, key)) {
             *ty
         } else {
             Self::insert_cold(
                 &mut self.archetypes,
                 &mut self.insert_map,
                 &mut self.remove_map,
-                TypeId::of::<B>(),
+                key,
                 B::insert,
                 meta.ty,
             )
@@ -207,7 +209,7 @@ impl World {
         archetypes: &mut Vec<Archetype>,
         insert_map: &mut IndexTypeIdMap<u16>,
         remove_map: &mut IndexTypeIdMap<u16>,
-        type_id: TypeId,
+        key: TypeId,
         insert: fn(&mut TypeMetadataSet),
         old_ty: u16,
     ) -> u16 {
@@ -216,8 +218,8 @@ impl World {
 
         let new_ty = Self::get_or_insert(archetypes, types);
 
-        insert_map.insert((old_ty, type_id), new_ty);
-        remove_map.insert((new_ty, type_id), old_ty);
+        insert_map.insert((old_ty, key), new_ty);
+        remove_map.insert((new_ty, key), old_ty);
 
         new_ty
     }
@@ -243,14 +245,16 @@ impl World {
         let meta = &self.entities[ent.id as usize];
         assert_eq!(ent.gen, meta.gen, "Entity is stale");
 
-        let new_ty = if let Some(ty) = self.remove_map.get(&(meta.ty, TypeId::of::<B>())) {
+        let key = TypeId::of::<B>();
+
+        let new_ty = if let Some(ty) = self.remove_map.get(&(meta.ty, key)) {
             *ty
         } else {
             Self::remove_cold(
                 &mut self.archetypes,
                 &mut self.insert_map,
                 &mut self.remove_map,
-                TypeId::of::<B>(),
+                key,
                 B::remove,
                 meta.ty,
             )?
@@ -274,7 +278,7 @@ impl World {
         archetypes: &mut Vec<Archetype>,
         insert_map: &mut IndexTypeIdMap<u16>,
         remove_map: &mut IndexTypeIdMap<u16>,
-        type_id: TypeId,
+        key: TypeId,
         remove: fn(&mut TypeMetadataSet) -> Option<()>,
         old_ty: u16,
     ) -> Option<u16> {
@@ -283,8 +287,8 @@ impl World {
 
         let new_ty = Self::get_or_insert(archetypes, types);
 
-        remove_map.insert((old_ty, type_id), new_ty);
-        insert_map.insert((new_ty, type_id), old_ty);
+        remove_map.insert((old_ty, key), new_ty);
+        insert_map.insert((new_ty, key), old_ty);
 
         Some(new_ty)
     }
@@ -316,13 +320,15 @@ impl World {
         let meta = &self.entities[ent.id as usize];
         assert_eq!(ent.gen, meta.gen, "Entity is stale");
 
-        let new_ty = if let Some(ty) = self.exchange_map.get(&(meta.ty, TypeId::of::<(B1, B2)>())) {
+        let key = TypeId::of::<(B1, B2)>();
+
+        let new_ty = if let Some(ty) = self.exchange_map.get(&(meta.ty, key)) {
             *ty
         } else {
             Self::exchange_cold(
                 &mut self.archetypes,
                 &mut self.exchange_map,
-                TypeId::of::<(B1, B2)>(),
+                key,
                 B1::remove,
                 B2::insert,
                 meta.ty,
@@ -335,7 +341,11 @@ impl World {
         unsafe {
             let old_comps = B1::read(&mut self.archetypes[old_ty as usize], old_idx);
 
-            let new_idx = self.move_(ent.id, old_ty, new_ty, old_idx);
+            let new_idx = if new_ty != old_ty {
+                self.move_(ent.id, old_ty, new_ty, old_idx)
+            } else {
+                old_idx
+            };
 
             new_comps.write(&mut self.archetypes[new_ty as usize], new_idx);
 
@@ -348,7 +358,7 @@ impl World {
     fn exchange_cold(
         archetypes: &mut Vec<Archetype>,
         exchange_map: &mut IndexTypeIdMap<u16>,
-        type_id: TypeId,
+        key: TypeId,
         remove: fn(&mut TypeMetadataSet) -> Option<()>,
         insert: fn(&mut TypeMetadataSet),
         old_ty: u16,
@@ -359,7 +369,7 @@ impl World {
 
         let new_ty = Self::get_or_insert(archetypes, types);
 
-        exchange_map.insert((old_ty, type_id), new_ty);
+        exchange_map.insert((old_ty, key), new_ty);
 
         Some(new_ty)
     }
@@ -382,6 +392,8 @@ impl World {
     }
 
     unsafe fn move_(&mut self, id: u32, old_ty: u16, new_ty: u16, old_idx: u32) -> u32 {
+        debug_assert_ne!(old_ty, new_ty);
+
         let old_archetype = &mut *self.archetypes.as_mut_ptr().add(old_ty as usize);
         let new_archetype = &mut *self.archetypes.as_mut_ptr().add(new_ty as usize);
 
@@ -717,11 +729,11 @@ struct IndexTypeIdHasher(u64);
 
 impl Hasher for IndexTypeIdHasher {
     fn write_u16(&mut self, val: u16) {
-        self.0 = 1 + val as u64;
+        self.0 = u64::from(val);
     }
 
     fn write_u64(&mut self, val: u64) {
-        self.0 = self.0.wrapping_mul(val);
+        self.0 ^= val;
     }
 
     fn write(&mut self, _val: &[u8]) {
@@ -740,11 +752,11 @@ struct IndexTagHasher(u64);
 
 impl Hasher for IndexTagHasher {
     fn write_u16(&mut self, val: u16) {
-        self.0 = val as u64;
+        self.0 = u64::from(val);
     }
 
     fn write_u32(&mut self, val: u32) {
-        self.0 = self.0 << 32 | val as u64;
+        self.0 |= u64::from(val) << 16;
     }
 
     fn write(&mut self, _val: &[u8]) {
@@ -760,6 +772,7 @@ impl Hasher for IndexTagHasher {
 mod tests {
     use super::*;
 
+    use std::hash::Hash;
     use std::mem::size_of;
 
     #[test]
@@ -924,6 +937,14 @@ mod tests {
         assert_eq!(world.remove_map[&(1, TypeId::of::<(i32,)>())], 2);
     }
 
+    #[test]
+    fn trival_exchange_does_not_create_aliasing_unique_references() {
+        let mut world = World::new();
+        let ent = world.alloc();
+        world.insert(ent, (true,));
+        world.exchange::<(bool,), _>(ent, (false,)).unwrap();
+    }
+
     #[test]
     fn insert_then_get() {
         let mut world = World::new();
@@ -1059,15 +1080,38 @@ mod tests {
         assert_eq!(*comp, 23);
     }
 
+    #[test]
+    fn index_type_id_yields_uniformly_distributed_lower_bits() {
+        let mut histogram = [0; 128];
+
+        for i in 0_u16..1024 {
+            for t in [
+                TypeId::of::<(i32,)>(),
+                TypeId::of::<(bool, f32)>(),
+                TypeId::of::<(i32, &str)>(),
+                TypeId::of::<(bool, &str, f64)>(),
+            ] {
+                let mut hasher = IndexTypeIdHasher::default();
+                (i, t).hash(&mut hasher);
+                let hash = hasher.finish();
+
+                histogram[hash as usize % histogram.len()] += 1;
+            }
+        }
+
+        for count in histogram {
+            assert_eq!(count, 1024 * 4 / histogram.len());
+        }
+    }
+
     #[test]
     fn index_tag_hasher_yields_uniformly_distributed_lower_bits() {
         let mut histogram = [0; 128];
 
-        for i in 0..1024 {
-            for j in 0..128 {
+        for i in 0_u16..1024 {
+            for j in 0_u32..128 {
                 let mut hasher = IndexTagHasher::default();
-                hasher.write_u32(i);
-                hasher.write_u32(j);
+                (i, j).hash(&mut hasher);
                 let hash = hasher.finish();
 
                 histogram[hash as usize % histogram.len()] += 1;
